; SKMF84.ASM - Straight Key Memory using the 16F84 PIC
;
; 1/12/2000     Chuck Olson, WB9KZY
;               Jackson Harbor Press
;               http://jacksonharbor.home.att.net
;               jacksonharbor@att.net
;
;       What does it do?
;
;       This program will use a 16f84 along with an external EEPROM chip to
;       create a straight key memory - it will record the actual length
;       of a dit, dah or space to a 2 ms (milli-second) precision - each byte of the
;       eeprom will be either a mark or space (flagged by the msb (most 
;       significant bit) - 1 = mark, 0 = space, so a stored byte of 
;       7FH or 01111111B is a space 254 ms long,
;       a stored byte of 81H or 10000001B is a 2 ms mark) - each element is 
;       at least 2 ms in length, normally they will be more than 20 ms since the
;       minimum dit length or space length will be 20 ms which equates to  
;       a top speed of 60 WPM.  Note that a stored byte of value 00H indicates 
;       the end of the message.  The playback speed can be varied
;       using the variable PIC RC clock. The unit will work
;       with either a straight key, a bug or an electronic keyer.  It will 
;       sense key down as a ground and key up as whatever - isolation diodes
;       can be used for non 5V inputs.  The unit does NOT include a
;       sidetone output.  There are 4 switch inputs which allow
;       for multiple memories.  When not playing back or recording a message
;       the program will loop - if the key input or the memory switch
;       inputs are pressed, the skm will start either playing or
;       recording messages or the output will "follow" the key input.
;       
;       The external EEPROM memory is optional - short messages can be
;       stored in the 16F84 internal RAM and the 16F84 internal EEPROM.
;
;       The code for driving the external eeprom is from Microchip AN515 - it
;       has been modified slightly.
;
;       Memory capacity - this will vary by the keying speed and the clock
;       setting of the PIC.  But for medium speeds where the dit, dah and
;       space elements are all less than 254 ms, each element will take one
;       byte of the eeprom.  Note that a 254 ms dah or char space would
;       indicate a keying speed of about 14 WPM.
;
;       For the word, PARIS, the element count would be a minimum of:
;       P  *--*  8 elements (2 dits, 2 dahs, 3 dit spaces and 1 char space)
;       A  *-    4           1       1       1                1
;       R  *-*   6           2       1       2                1
;       I  **    4           1                                1
;       S  ***   6           3               2                1
;       Total:  28 bytes of eeprom would be needed to hold PARIS (minimum)
;       
;       memory  type            size (bytes/elements)  character capacity
;       1       ram                     54             ? depends on the
;       2       internal eeprom         64              speed during
;       3       external eeprom       1024              record and on the
;       4       external eeprom       1024              characters recorded
;
; For example:  DE WB9KZY (54 code elements) fits perfectly into mem 1 (ram)
;               
;
;       Future Enhancements ? :
;       1) add a sleep mode to minimize current ?
;       2) add a message kill from the key input (instead of grounding mclr)?
;       3) add a sidetone output - would allow "dead" recording ?
;       4) add an end of message LED indicator &/or other LED indicators ?
;       5) eliminate internal RAM message and replace with external eeprom ?
;       6) add 2 (or more) memories using rb1,2,3 or others for switch inputs?
;       7) add manual pausing of message play by pressing a memory switch ?
;       8) add a test for the end of the external eeprom memories during
;          record and/or play ? (now they will run-over the boundry into
;          the other memory.
;       9) eliminate the trailing space at the end of a memory record ?
;          (this is the space recorded after you finish sending and before
;           you get a chance to press the memory button to end the recording)
;
;
;PIC
;pin    name    function/connection
;---    ----    -----------------------------------------------
;  1    ra2     SCL (pin 6 of 24LC16 eeprom)
;  2    ra3     SDA (pin 5 of 24LC16 eeprom)
;  3    ra4     no connect
;  4    MCLR    pulled up to +5V with a 10K resistor
;  5    VSS     Ground 
;  6    rb0     key input - use the internal portb pullup to +5V
;  7    rb1     no connect
;  8    rb2     no connect
;  9    rb3     no connect
; 10    rb4     memory switch 1 - use the internal portb pullup to +5V
; 11    rb5     memory switch 2 - use the internal portb pullup to +5V
; 12    rb6     memory switch 3 - use the internal portb pullup to +5V
; 13    rb7     memory switch 4 - use the internal portb pullup to +5V
; 14    VDD     +5V
; 15    osc2    no connect - can be used to measure PIC clock/4 w/counter
; 16    osc1    connected to RC clock circuit
; 17    ra0     no connect
; 18    ra1     keying output, follows state of pin 6 or memory play


	LIST P=16F84, R=DEC 
	INCLUDE "p16F84.inc"
	__FUSES _PWRTE_OFF & _CP_OFF & _WDT_OFF & _RC_OSC
; powerup timer off - not sure why.
; copy protect is off
; watchdog timer is off 
; the oscillator uses an external RC pair


; constants
freq    equ     d'165'  ;tooms delay loop constant
eemsg   equ     H'00'   ;start of internal eeprom message memory
eemsgen equ     H'3F'   ;end of internal eeprom message memory


; memory locations
temp    equ     H'16'   ;temporary location used in halfcyc
twoms   equ     H'17'   ;counter location for 2ms increments
eeaddr  equ     H'18'   ;internal eeprom address holder - needed for write
mess    equ     H'19'   ;start of ram message memory
messend equ     H'4F'   ;end of ram message memory


;other bit definitions
zero    equ     H'02'   ;zero flag in STATUS
sw1     equ     H'07'   ;switch 1 = PORTB bit 7
sw2     equ     H'06'   ;switch 2 = PORTB bit 6
sw3     equ     h'05'   ;switch 3 = PORTB bit 5
sw4     equ     h'04'   ;switch 4 = PORTB bit 4
in      equ     H'00'   ;in = input pin PORTB bit 0
out     equ     H'01'   ;out = output pin PORTA bit 1


	ORG     0x00
start   goto    init    ;startup routine
	

	ORG     0x04
rupt    goto    init    ;interrupt routine


; tooms - a routine to wait 2 ms at 4 MHz PIC clock (1 us per instr cycle)
;delay: 2(call) + 3(setup) + ((freq-1) *12) + 11(last decfsz) + 2(nop) + 2(return)
;     this reduces to: 20 + 165 * 12 = 2000 us or 2 ms
tooms   movlw   freq    ;1    3 load up the 1/2 cycle delay constant
	movwf   temp    ;1    4 store it in the temporary location
loop2   goto    l3      ;2    waste 2 cycles
l3      goto    l4      ;2    waste another 2 cycles
l4      goto    l5      ;2    waste yet another 2 cycles
l5      goto    l6      ;2    waste a fourth set of 2 cycles
l6      nop             ;1    waste just 1 cycle for a total addition of 9
	decfsz  temp,1  ;1    decrement the temp location - is it 0 ?
	goto    loop2   ;2    no, keep decrementing
	;yes, return to the calling routine
	goto    end2ms  ;2    waste 2 cycles
end2ms  return          ;2    end of 2 ms wait routine


;    deboun - a routine to wait 20 ms
; Why do a brute force 10 calls of tooms rather than a smaller loop?
; Because, it's only 3 words longer (my loop was 7 words) - also it
;       doesn't require a loop counter which saves a byte of RAM for 
;       the message - also the accuracy is better - total debounce delay
;       is 20 ms plus 4 us (2 us for call of deboun and 2 us for return)
;       at a PIC clock speed of 4 MHz.
deboun  call    tooms   ;  2 ms 
	call    tooms   ;  4 ms 
	call    tooms   ;  6 ms 
	call    tooms   ;  8 ms 
	call    tooms   ; 10 ms 
	call    tooms   ; 12 ms 
	call    tooms   ; 14 ms 
	call    tooms   ; 16 ms 
	call    tooms   ; 18 ms 
	call    tooms   ; 20 ms total
dbdun   return          ;end of the debounce routine


charsp  movlw   d'120'  ;a routine to send a 15 wpm character space (3 dit spaces)
	goto    os      ;enter outdit just after the dit space is loaded


outdah  movlw   d'120'  ;a routine to send a 15 wpm dah
	goto    od      ;enter outdit just after the dit length is loaded


outdit  movlw   d'40'  ;15 wpm dit
od      movwf   twoms   ;save it in twoms 
	bsf     PORTA,out;no, it's a mark so set the output high
ti      call    tooms   ;wait 2 ms
	decfsz  twoms,1 ;decrement the 2 ms counter - is it zero?
	goto    ti      ;no, keep looping
	bcf     PORTA,out;clear the output
	movlw   d'40'   ;15 wpm dit space
os      movwf   twoms   ;save it in twoms 
ti2     call    tooms   ;wait 2 ms
	decfsz  twoms,1 ;decrement the 2 ms counter - is it zero?
	goto    ti2     ;no, keep looping
	return          ;end of the outdit/dah routines


; init - the initialization power up entry point of the program
init    bsf     STATUS,5        ;set the bank bit to 1
	movlw   b'11110001'     ; inputs on PORTB bits 0, 4, 5, 6, 7
	movwf   TRISB           ; set the tri-state register
	movlw   b'00000000'     ; all outputs on PORTA
	movwf   TRISA           ; set the tri-state register
	movlw   b'00000000'     ; no option bits are relevant xcept PORTB p/u
	movwf   OPTION_REG      ; set the option register
	bcf     STATUS,5        ;set the bank bit to 0
	bcf     INTCON,GIE      ;disable interrupts
;removed the following to allow using mclr (pin 4) as a message kill pin
;       if pin 4 is grounded temporarily, it will reset the PIC and
;       thus kill any message currently being sent.
;        clrf    mess            ;initialize ram msg to no characters
	bcf     PORTA,out       ;set the true output off (inverse of pin 1)
	call    deboun          ;No, wait for a bit after powerup        


mainlup ;check for switch presses and branch if one occurs
	btfss   PORTB, in       ;see if the input is active
	goto    check           ;yes, goto the input check
	btfss   PORTB, sw1      ;no, see if the sw1 switch is pressed
	goto    memory          ;yes, go to the memory routine
	btfss   PORTB, sw2      ;no, see if sw2 switch is pressed
	goto    memory2         ;yes, enter the 2nd memory routine
	btfss   PORTB, sw3      ;no, see if sw3 switch is pressed
	goto    memory3         ;yes, enter the 3rd memory routine
	btfss   PORTB, sw4      ;no, see if sw4 switch is pressed
	goto    memory4         ;yes, enter the 4th memory routine
	goto    mainlup         ;keep looping


check   bsf     PORTA, out      ;turn on the output
	call    deboun          ;wait 20 ms to debounce the switch
intst   btfss   PORTB, in       ;see if the input is still low
	goto    intst           ;yes, loop back
	bcf     PORTA, out      ;no, turn off the output
	call    deboun          ;wait 20 ms to debounce the switch
	btfss   PORTB, in       ;is the input is low again??
	goto    check           ;yes, loop back
	goto    mainlup         ;no, go back and check the switches


; memory - a routine to wait until the dit paddle is hit OR the memory
;       switch is released.
; press and release sw1 (PORTB bit 3) to play message 1
; press and hold sw1 (wait for q mark) to record message 1
memory  clrf    twoms           ;clear a key down counter
memp1   call    deboun          ;wait for 20 ms
	incf    twoms,1         ;increment the key down counter
	movlw   d'100'          ;load up a constant of 100
	subwf   twoms,0         ;see if you've waited 2 seconds (100 * 20ms)
	btfsc   STATUS,zero     ;if the z flag is set you've waited 2 sec
	goto    record          ;yes, goto ram record routine   
	btfss   PORTB, sw1      ;test sw1 - is it on?
	goto    memp1           ;it is, keep looping for now
	;it isn't 
	call    deboun          ;wait for 20 ms
	;play message 1
; play - a routine to demo a play of length based recording in RAM
play    btfss   PORTB, sw1      ;is sw1 still pressed?
	goto    play            ;yes, loop until release
	;no, start playing the RAM contents
	movlw   mess            ;start of the message stored in RAM    
	movwf   FSR             ;save it into the indirect pointer
ldele   movf    INDF,0          ;get an element and length
	movwf   twoms           ;save it in twoms 
	movf    twoms,1         ;test the byte
	btfsc   STATUS,zero     ;is this the end of the message (0)??
	goto    endplay         ;yes, end the play
	;no, keep sending
	btfss   twoms,7         ;see if the element is a space (bit 7 =0)??
	goto    sp              ;yes, bit 7 is a 0 so send a space
mark    bsf     PORTA,out       ;no, it's a mark so set the output high
	bcf     twoms,7         ;clear the mark bit
	goto    timeit          ;branch to the timout routine
sp      bcf     PORTA,out       ;set the output low
timeit  call    tooms           ;wait 2 ms
	decfsz  twoms,1         ;decrement the 2 ms counter - is it zero?
	goto    timeit          ;no, keep looping
	movlw   messend         ;lets look at the element counter
	subwf   FSR,W           ;is the element counter 
	btfsc   STATUS,Z        ;      less than the message end address?
	goto    endplay         ;no, they are equal - bail out.
	incf    FSR,1           ;yes, increment the element counter
	goto    ldele           ;get the next element
endplay bcf     PORTA,out       ;clear the output
	goto    mainlup         ;go back to the main loop


; qmark - a routine to send a question mark 
qmark   call  outdit    ;send a q mark 
	call  outdit
	call  outdah
	call  outdah
	call  outdit
	call  outdit
	call  charsp    ;       + a character space
	return          ;end of the qmark routine


; record - a routine to record a short message into ram
record  call    qmark           ;send a q mark
	btfss   PORTB, sw1      ; see if sw1 is pressed
	goto    record          ;yes, keep sending the q mark until sw1 released
	movlw   mess            ;load up the start of the message buffer
	movwf   FSR             ;save the callsign buffer pointer
check2  btfss   PORTB, sw1      ; see if sw1 is pressed
	goto    vvait           ;yes, bailout of record without recording 
	btfsc   PORTB, in       ;see if the input is active
	goto    check2          ;it isn't active, wait for a switch closure
check3  bsf     PORTA, out      ;it is active, turn on the output
	call    deboun          ;wait 20 ms
	movlw   d'138'  ;load up a 20 ms count (10 x 2 ms) + 128 to set bit 7
	movwf   twoms           ;store it in the 2 ms counter
intst2  btfsc   PORTB, in       ;see if the input is still low
	goto    keyup           ;the key is up
	call    tooms           ;wait 2 ms
	incfsz  twoms  ,1       ;increment the 2 ms counter
	goto    intst2          ;keep looping until the key is up
	;the counter has rolled over, put in a full byte and keep looping
	movlw   d'255'          ;load all ones into w
	movwf   INDF            ;save it in memory
	incf    FSR,1           ;increment the memory pointer
	movlw   messend         ;lets look at the element counter
	subwf   FSR,W           ;is the element counter 
	btfsc   STATUS,Z        ;        less than the message end address?
	goto    bailo           ;no, they are equal - bail out.
	movlw   d'128'          ;load up 128 to set bit 7 for the mark.
	movwf   twoms           ;store it in the 2 ms counter
	goto    intst2          ;keep looping until the key is released
keyup   bcf     PORTA, out      ;turn off the output
	bsf     twoms  ,7       ;set the high order bit a 1 to signify a mark
	movf    twoms  ,0       ;load the key down count
	movwf   INDF            ;save it in memory
	incf    FSR,1           ;increment the memory pointer
	movlw   messend         ;lets look at the element counter
	subwf   FSR,W           ;is the element counter 
	btfsc   STATUS,Z        ;        less than the message end address?
	goto    bailo           ;no, they are equal - bail out.
	call    deboun          ;wait 20 ms
	movlw   d'138'  ;load up a 20 ms count (10 x 2 ms) + 128 to set bit 7
	movwf   twoms           ;store it in the 2 ms counter
intst3  btfss   PORTB, in       ;see if the input is still high
	goto    keydwn          ;no, the key is down
	btfss   PORTB, sw1      ; yes, see if sw1 is pressed
	goto    bailo           ;yes, bailout of record
	call    tooms           ;wait 2 ms
	incfsz  twoms  ,1       ;increment the 2 ms counter
	goto    intst3          ;keep looping until the key is pressed
	;the counter has rolled over, put in a full byte and keep looping
	movlw   d'127'          ;load all ones into w (except bit 7)
	movwf   INDF            ;save it in memory
	incf    FSR,1           ;increment the memory pointer
	movlw   messend         ;lets look at the element counter
	subwf   FSR,W           ;is the element counter 
	btfsc   STATUS,Z        ;        less than the message end address?
	goto    bailo           ;no, they are equal - bail out.
	movlw   d'128'          ;load up 128 to set bit 7 for the mark.
	movwf   twoms           ;store it in the 2 ms counter
	goto    intst3  ;keep looping until sw1 or key is pressed
keydwn  ;bsf     PORTA, out     ;turn on the output
	bcf     twoms  ,7       ;set the high order bit a 1 to signify a mark
	movf    twoms  ,0       ;load the key down count
	movwf   INDF            ;save it in memory
	incf    FSR,1           ;increment the memory pointer
	movlw   messend         ;lets look at the element counter
	subwf   FSR,W           ;is the element counter 
	btfss   STATUS,Z        ;       less than the message end address?
	goto    check3          ;start the next mark element
bailo   movlw   0               ;load a 0 for the end of message byte
	movwf   INDF            ;save the end of message byte
vvait   call    deboun          ;wait a while
	btfss   PORTB,sw1       ;test sw1
	goto    vvait           ;wait until sw1 is released then
	goto    mainlup         ;end of the record routine
	

; memory2 - a routine to either play or enter record for mem2 (internal eeprom)
; press and release sw2 to play message 2
; press and hold sw2 (wait for q mark) to record message 2
memory2 clrf    twoms           ;clear a key down counter
memp2   call    deboun          ;wait for 20 ms
	incf    twoms,1         ;increment the key down counter
	movlw   d'100'          ;load up a constant of 100
	subwf   twoms,0         ;see if you've waited 2 seconds (100 * 20ms)
	btfsc   STATUS,zero     ;if the z flag is set you've waited 2 sec
	goto    record2         ;yes, goto internal EEPROM record routine   
	btfss   PORTB, sw2      ;test sw2 - is it on?
	goto    memp2           ;it is, keep looping for now
	;it isn't 
	call    deboun          ;wait for 20 ms
	;play message 2
; play2 - a routine to demo a play of length based recording in int eeprom
play2   btfss   PORTB, sw2      ;is sw2 still pressed?
	goto    play2           ;yes, loop until release
	;no, start playing the EEPROM contents
	clrf    EEADR   ;start of the message stored in eeprom is zero
ldele2  bsf     STATUS,5        ;set the bank bit to 1 for EECON1 set
	bsf     EECON1,RD       ;setup internal EEPROM for read mode
	bcf     STATUS,5        ;reset the bank bit to 0
	movf    EEDATA,0        ;load up w with the EEPROM data byte
	movwf   twoms           ;save it in twoms 
	movf    twoms,1         ;test the byte
	btfsc   STATUS,zero     ;is this the end of the message (0)??
	goto    endply2         ;yes, end the play
	;no, keep sending
	btfss   twoms,7         ;see if the element is a space (bit 7 =0)??
	goto    sp2             ;yes, bit 7 is a 0 so send a space
mark2   bsf     PORTA,out       ;no, it's a mark so set the output high
	bcf     twoms,7         ;clear the mark bit
	goto    timeit2         ;branch to the timout routine
sp2     bcf     PORTA,out       ;set the output low
timeit2 call    tooms           ;wait 2 ms
	decfsz  twoms,1         ;decrement the 2 ms counter - is it zero?
	goto    timeit2         ;no, keep looping
	movlw   eemsgen         ;lets look at the element counter
	subwf   EEADR,W         ;is the element counter 
	btfsc   STATUS,Z        ;        less than the message end address?
	goto    endply2         ;no, they are equal - bail out.
	incf    EEADR,1         ;increment the element counter
	goto    ldele2          ;yes, get the next element
endply2 bcf     PORTA,out       ;clear the output
	goto    mainlup         ;head back to the main loop


;  record2 - a routine to record a short message into internal EEPROM
; This routine is very similar to record except that the data is stored in
; the 16F84 eeprom which requires slightly different routines than ram.
record2 call    qmark           ;send a q mark
	btfss   PORTB, sw2      ; see if sw2 is pressed
	goto    record2         ;yes, keep sending the q mark until sw2 released
	clrf    eeaddr          ;no, load up the start of the message buffer
check4  btfss   PORTB, sw2      ; see if sw2 is pressed
	goto    vvait2          ;yes, bailout of record2 without recording
	btfsc   PORTB, in       ;see if the input is active
	goto    check4          ;it isn't active, wait for a switch closure
check5 ;the key is down - start recording!
	bsf     PORTA, out      ;it is active, turn on the output
	call    deboun          ;wait 20 ms
	movlw   d'138'  ;load up a 20 ms count (10 x 2 ms) + 128 to set bit 7
	movwf   twoms           ;store it in the 2 ms counter
intst4  btfsc   PORTB, in       ;see if the input is still low
	goto    keyup2          ;the key is up
	call    tooms           ;wait 2 ms
	incfsz  twoms  ,1       ;increment the 2 ms counter
	goto    intst4          ;keep looping until the key is up
	;the counter has rolled over, put in a full byte and keep looping
	movlw   d'255'          ;load all ones into w
	movwf   EEDATA          ;save it in the EEPROM data buffer
	call    wreep           ;write the data
	incf    eeaddr,1        ;increment the element counter
	movlw   eemsgen         ;lets look at the element counter
	subwf   eeaddr,W           ;is the element counter 
	btfsc   STATUS,Z        ;        less than the message end address?
	goto    bailo2          ;no, they are equal - bail out.
	movlw   d'128'          ;load up 128 to set bit 7 for the mark.
	movwf   twoms           ;store it in the 2 ms counter
	goto    intst4          ;keep looping until the key is released
keyup2  bcf     PORTA, out      ;turn off the output
	bsf     twoms  ,7       ;set the high order bit a 1 to signify a mark
	movf    twoms  ,0       ;load the key down count
	movwf   EEDATA          ;save it in the EEPROM data buffer
	call    wreep           ;write the data
	incf    eeaddr,1        ;increment the element counter
	movlw   eemsgen         ;lets look at the element counter
	subwf   eeaddr,W           ;is the element counter 
	btfsc   STATUS,Z        ;        less than the message end address?
	goto    bailo2          ;no, they are equal - bail out.
	call    deboun          ;wait 20 ms
	movlw   d'138'          ;load up a 20 ms count (10 x 2 ms) + 128
	;why do we add 128 above? - to allow the counter to rollover to 0
	; when we exceed a space of 256 ms.
	movwf   twoms           ;store it in the 2 ms counter
intst5  btfss   PORTB, in       ;see if the input is still high
	goto    keydwn2         ;no, the key is down
	btfss   PORTB, sw2      ; yes, see if sw2 is pressed
	goto    bailo2          ;yes, bailout of record
	call    tooms           ;wait 2 ms
	incfsz  twoms  ,1       ;increment the 2 ms counter
	goto    intst5          ;keep looping until the key is pressed
	;the counter has rolled over, put in a full byte and keep looping
	movlw   d'127'          ;load all ones into w (except bit 7)
	movwf   EEDATA          ;save it in the EEPROM data buffer
	call    wreep           ;write the data
	incf    eeaddr,1        ;increment the element counter
	movlw   eemsgen         ;lets look at the element counter
	subwf   eeaddr,W           ;is the element counter 
	btfsc   STATUS,Z        ;        less than the message end address?
	goto    bailo2          ;no, they are equal - bail out.
	movlw   d'128'          ;load up 128 to set bit 7 for the mark.
	movwf   twoms           ;store it in the 2 ms counter
	goto    intst5          ;keep looping until sw1 or key is pressed
keydwn2 bcf     twoms  ,7       ;clear the high order bit a 1 to signify a space
	movf    twoms  ,0       ;load the key down count
	movwf   EEDATA          ;save it in the EEPROM data buffer
	call    wreep           ;write the data
	incf    eeaddr,1        ;increment the element counter
	movlw   eemsgen         ;lets look at the element counter
	subwf   eeaddr,W           ;is the element counter 
	btfsc   STATUS,Z        ;        less than the message end address?
	goto    bailo2          ;no, they are equal - bail out.
	goto    check5          ;start the next mark element
bailo2  movlw   H'00'           ;load a 0 for the end of message byte
	movwf   EEDATA          ;save it in the EEPROM data buffer
	call    wreep           ;write the data
vvait2  call    deboun          ;wait a while
	btfss   PORTB,sw2       ;test sw2
	goto    vvait2          ;wait until sw2 is released then
	goto    mainlup         ;end of the record2 routine
	
	
;    wreep - a routine to write the data in EEDATA into the address EEADR
; write sequence from the 16F84 spec - for the internal 16f84 eeprom.
; Why use the eeaddr memory location to hold the internal eeprom address
;       rather than using EEADR?
; Because EEADR appears to be reset to 0 after each write to the internal
;       eeprom, so it can't be used as an address pointer for the record2 
;       routine - this doesn't happen with the read from the internal eeprom.
wreep   movf    eeaddr,0        ;load up the internal eeprom address
	movwf   EEADR           ;store is in the EEPROM address buffer
	bsf     STATUS,5        ;set the bank bit to 1
checkit btfsc   EECON1,WR       ;see if the last write is complete
	goto    checkit         ;no, check it again
	bsf     EECON1,WREN     ;enable write
	movlw   H'55'           ;load up a 55h per the 16f84 spec
	movwf   EECON2          ;write the 55h 
	movlw   H'AA'           ;load up AAh per the 16f84 spec
	movwf   EECON2          ;write the AAh 
	bsf     EECON1,WR       ;set the WR bit and thus begin writing to eeprom
	bcf     EECON1,WREN     ;disable write
	bcf     STATUS,5        ;reset the bank bit to 0
	return                  ;end of wreep routine


;memory3 - a routine to either play or enter record for mem3 (external eeprom)
; press and release sw3 to play message 3
; press and hold sw3 (wait for q mark) to record message 3
memory3 clrf    twoms           ;clear a key down counter
memp3   call    deboun          ;wait for 20 ms
	incf    twoms,1         ;increment the key down counter
	movlw   d'100'          ;load up a constant of 100
	subwf   twoms,0         ;see if you've waited 2 seconds (100 * 20ms)
	btfsc   STATUS,zero     ;if the z flag is set you've waited 2 sec
	goto    record3         ;yes, goto internal EEPROM record routine   
	btfss   PORTB, sw3      ;test sw3 - is it on?
	goto    memp3           ;it is, keep looping for now
	;it isn't 
	call    deboun          ;wait for 20 ms
	;play message 3
; play3 - a routine to demo a play of length based recording in int eeprom
play3   btfss   PORTB, sw3      ;is sw3 still pressed?
	goto    play3           ;yes, loop until release
	;no, start playing the EEPROM contents
	clrf    ADDR            ;start of the message stored in eeprom is zero
	clrf    eepage          ;init the start of the message buffer page
ldele3  call    RDBYTE          ;get a byte
	movf    DATAI,0         ;load up w with the EEPROM data byte
	movwf   twoms           ;save it in twoms 
	movf    twoms,1         ;test the byte
	btfsc   STATUS,zero     ;is this the end of the message (0)??
	goto    endply3         ;yes, end the play
	;no, keep sending
	btfss   twoms,7         ;see if the element is a space (bit 7 =0)??
	goto    sp3             ;yes, bit 7 is a 0 so send a space
mark3   bsf     PORTA,out       ;no, it's a mark so set the output high
	bcf     twoms,7         ;clear the mark bit
	goto    timeit3         ;branch to the timout routine
sp3     bcf     PORTA,out       ;set the output low
timeit3 call    tooms           ;wait 2 ms
	decfsz  twoms,1         ;decrement the 2 ms counter - is it zero?
	goto    timeit3         ;no, keep looping
	;add an element and/or page counter check here ?
	incfsz  ADDR,1          ;increment the element counter
	goto    lt256a          ;it's not zero, leave the page alone
	incf    eepage,1        ;it's zero, increment the next page 
lt256a       
	goto    ldele3          ;yes, get the next element
endply3 bcf     PORTA,out       ;clear the output
	goto    mainlup         ;head back to the main loop


;  record3 - a routine to record a short message into external EEPROM
; The data is stored in the 24LC16 external eeprom which requires totally 
; different routines than the internal eeprom of the 16f84.
record3 call    qmark           ;send a q mark
	btfss   PORTB, sw3      ; see if sw3 is pressed
	goto    record3         ;yes, keep sending the q mark until sw3 released
	clrf    ADDR            ;no, load up the start of the message buffer
	clrf    eepage          ;init the start of the message buffer page
check6  btfss   PORTB, sw3      ; see if sw3 is pressed
	goto    vvait3          ;yes, bailout of record3 without recording
	btfsc   PORTB, in       ;see if the input is active
	goto    check6          ;it isn't active, wait for a switch closure
check7 ;the key is down - start recording!
	bsf     PORTA, out      ;it is active, turn on the output
	call    deboun          ;wait 20 ms
	movlw   d'138'  ;load up a 20 ms count (10 x 2 ms) + 128 to set bit 7
	movwf   twoms           ;store it in the 2 ms counter
intst6  btfsc   PORTB, in       ;see if the input is still low
	goto    keyup3          ;the key is up
	call    tooms           ;wait 2 ms
	incfsz  twoms  ,1       ;increment the 2 ms counter
	goto    intst6          ;keep looping until the key is up
	;the counter has rolled over, put in a full byte and keep looping
	movlw   d'255'          ;load all ones into w
	call    savninc         ;save the byte to eeprom and inc counters
	;add an element and/or page counter check here ?
	movlw   d'128'          ;load up 128 to set bit 7 for the mark.
	movwf   twoms           ;store it in the 2 ms counter
	goto    intst6          ;keep looping until the key is released
keyup3  bcf     PORTA, out      ;turn off the output
	bsf     twoms  ,7       ;set the high order bit a 1 to signify a mark
	movf    twoms  ,0       ;load the key down count
	call    savninc         ;save the byte to eeprom and inc counters
	;add an element and/or page counter check here ?
	call    deboun          ;wait 20 ms
	movlw   d'138'          ;load up a 20 ms count (10 x 2 ms) + 128
	;why do we add 128 above? - to allow the counter to rollover to 0
	; when we exceed a space of 256 ms.
	movwf   twoms           ;store it in the 2 ms counter
intst7  btfss   PORTB, in       ;see if the input is still high
	goto    keydwn3         ;no, the key is down
	btfss   PORTB, sw3      ; yes, see if sw3 is pressed
	goto    bailo3          ;yes, bailout of record
	;added following 2 lines to allow recording of memory4
	btfss   PORTB, sw4      ; yes, see if sw4 is pressed
	goto    bailo3          ;yes, bailout of record
	call    tooms           ;wait 2 ms
	incfsz  twoms  ,1       ;increment the 2 ms counter
	goto    intst7          ;keep looping until the key is pressed
	;the counter has rolled over, put in a full byte and keep looping
	movlw   d'127'          ;load all ones into w (except bit 7)
	call    savninc         ;save the byte to eeprom and inc counters
	;add an element and/or page counter check here ?
	movlw   d'128'          ;load up 128 to set bit 7 for the mark.
	movwf   twoms           ;store it in the 2 ms counter
	goto    intst7          ;keep looping until sw1 or key is pressed
keydwn3 bcf     twoms  ,7       ;clear the high order bit a 1 to signify a space
	movf    twoms  ,0       ;load the key down count
	call    savninc         ;save the byte to eeprom and inc counters
	;add an element and/or page counter check here ?
	goto    check7          ;start the next mark element
bailo3  movlw   H'00'           ;load a 0 for the end of message byte
	movwf   DATAO           ;move it to the buffer
	call    WRBYTE          ;write the byte
vvait3  call    deboun          ;wait a while
	btfss   PORTB,sw3       ;test sw3
	goto    vvait3          ;wait until sw3 is released then
	;added following 2 lines to allow recording of memory 4
	btfss   PORTB,sw4       ;test sw4
	goto    vvait3          ;wait until sw4 is released then
	goto    mainlup         ;end of the record3 routine


; savninc - a routine to save the recorded byte to eeprom and then increment
;       the eeprom address counter and if needed, the eeprom page counter
savninc movwf   DATAO           ;move it to the buffer
	call    WRBYTE          ;write the byte
	incfsz  ADDR,1          ;increment the element counter
	goto    lt256           ;it's not zero, leave the page alone
	incf    eepage,1        ;it's zero, increment the next page 
lt256   return


;memory4 - a routine to either play or enter record for mem4 (external eeprom)
; press and release sw4 to play message 4
; press and hold sw4 (wait for q mark) to record message 4
; Since the code for mem4 is so similar to mem3, branched into the play3
; routine to save a little code.
memory4 clrf    twoms           ;clear a key down counter
memp4   call    deboun          ;wait for 20 ms
	incf    twoms,1         ;increment the key down counter
	movlw   d'100'          ;load up a constant of 100
	subwf   twoms,0         ;see if you've waited 2 seconds (100 * 20ms)
	btfsc   STATUS,zero     ;if the z flag is set you've waited 2 sec
	goto    record4         ;yes, goto internal EEPROM record routine   
	btfss   PORTB, sw4      ;test sw3 - is it on?
	goto    memp4           ;it is, keep looping for now
	;it isn't 
	call    deboun          ;wait for 20 ms
	;play message 4
; play4 - a routine to demo a play of length based recording in int eeprom
play4   btfss   PORTB, sw4      ;is sw4 still pressed?
	goto    play4           ;yes, loop until release
	;no, start playing the EEPROM contents
	clrf    ADDR            ;start of the message stored in eeprom is zero
	movlw   4               ;load up the 5th of 8 pages of external eeprom
	movwf   eepage          ;init the start of the message buffer page
	goto    ldele3          ;branch into the play3 routine 


;  record4 - a routine to record a short message into external EEPROM
; The data is stored in the 24LC16 external eeprom which requires totally 
; different routines than the internal eeprom of the 16f84.
; Since the code for mem4 is so similar to mem3, branched into the record3 
; routine to save a little code.
record4 call    qmark           ;send a q mark
	btfss   PORTB, sw4      ; see if sw3 is pressed
	goto    record4         ;yes, keep sending the q mark until sw3 released
	clrf    ADDR            ;no, load up the start of the message buffer
	movlw   4               ;load up the 5th of 8 pages of external eeprom
	movwf   eepage          ;init the start of the message buffer page
check8  btfss   PORTB, sw4      ; see if sw3 is pressed
	goto    vvait3          ;yes, bailout of record without recording
	btfsc   PORTB, in       ;see if the input is active
	goto    check8          ;it isn't active, wait for a switch closure
	goto    check7          ;branch into the record3 routine


; the following code is from Microchip AN515
;*****************************************************************************
;**  Two wire/I2C Bus READ/WRITE Sample Routines of Microchip's
;**  24Cxx / 85Cxx serial CMOS EEPROM interfacing to a 
;**  PIC16C54 8-bit CMOS single chip microcomputer
;**  Revsied Version 2.0 (4/2/92).
;**
;**     Part use = PIC16C54-XT/JW                                           
;**  Note:  1) All timings are based on a reference crystal frequency of 2MHz
;**            which is equivalent to an instruction cycle time of 2 usec.
;**         2) Address and literal values are read in octal unless otherwise
;**            specified.
;
;
;       Program:          I2CBUS.ASM 
;       Revision Date:   
;                         12-12-95      Compatibility with MPASMWIN 1.30
;
;********************************************************************** 
;
;*****************************************************************************
;
;-----------------------------------------------------------------------------
;       Files Assignment
;-----------------------------------------------------------------------------
;
;
FLAG    EQU     H'0C'           ; Common flag bits register
EEPROM  EQU     H'0D'           ; Bit buffer
ERCODE  EQU     H'0E'           ; Error code (to indicate bus status)
ADDR    EQU     H'0F'           ; Address register
DATAI   EQU     H'10'           ; Stored data input register
DATAO   EQU     H'11'           ; Stored data output register
; substituted the memory location, eepage, instead of SLAVE
eepage  equ     H'12'           ; 256 byte page in memory
TXBUF   EQU     H'13'           ; TX buffer
RXBUF   EQU     H'14'           ; RX buffer
COUNT   EQU     H'15'           ; Bit counter
;as far as I can tell, TIMER0 and TIMER1 aren't used, so they were deleted               
;
;
;-----------------------------------------------------------------------------
;                     Bit Assignments
;-----------------------------------------------------------------------------
;
; Status bits
;
Z       EQU     2
C       EQU     0
;
; FLAG Bits
;
ERR1    EQU     0               ; Error flag
;
; Instruction Destination Bits
;
F       EQU     1
W       EQU     0
;
; EEPROM Bits
;
DI      EQU     3               ; EEPROM input
DO      EQU     2               ; EEPROM output
;
; I2C Device Bits
;
SDA     EQU     3               ; RA3, data in/out
SCL     EQU     2               ; RA2, serial clock
;
; Constants
;
;END FILES/BITS EQUATE
; PAGE
;
;-----------------------------------------------------------------------------
;       Two wire/I2C - CPU communication error status table and subroutine
;-----------------------------------------------------------------------------
;  input  :     W-reg   = error code
;  output :     ERCODE  = error code
;           FLAG(ERR1) = 1
;
;         code          error status mode
;       -------         ------------------------------------------------------
;           1   :       SCL locked low by device (bus is still busy)
;           2   :       SDA locked low by device (bus is still busy)
;           3   :       No acknowledge from device (no handshake)
;           4   :       SDA bus not released for master to generate STOP bit
;-----------------------------------------------------------------------------
;
;Subroutine to identify the status of the serial clock (SCL) and serial data
;(SDA) condition according to the error status table. Codes generated are
;useful for bus/device diagnosis.
;
ERR
	BTFSS   FLAG,ERR1       ; Remain as first error encountered
	MOVWF   ERCODE          ; Save error code
	BSF     FLAG,ERR1       ; Set error flag
	RETLW   0
;
;-----------------------------------------------------------------------------
;       START bus communication routine
;-----------------------------------------------------------------------------
;       input   : none
;       output  : initialize bus communication
;-----------------------------------------------------------------------------
;
;Generate START bit (SCL is high while SDA goes from high to low transition)
;and check status of the serial clock.
BSTART
	MOVLW   B'00000000'     ; Put SCL, SDA line in output state
	bsf     STATUS,5        ;set the bank bit to 1
	movwf   TRISA           ; set the tri-state register
	bcf     STATUS,5        ;clear the bank bit to 0
;***************************************************************************
	bsf     PORTA,SDA          ;make sure sda is high
;***************************************************************************
	BSF     PORTA,SCL          ; Set clock high
	MOVLW   1               ; Ready error status code 1
	BTFSS   PORTA,SCL          ; Locked?
	CALL    ERR             ; SCL locked low by device
	BCF     PORTA,SDA          ; SDA goes low during SCL high
	NOP                     ; Timing adjustment
	NOP
	NOP
	BCF     PORTA,SCL          ; Start clock train
	RETLW   0
;
;END SUB
; PAGE
;
;-----------------------------------------------------------------------------
;       STOP bus communication routine
;-----------------------------------------------------------------------------
;       Input   :       None
;       Output  :       Bus communication, STOP condition
;-----------------------------------------------------------------------------
;
;Generate STOP bit (SDA goes from low to high during SCL high state)
;and check bus conditions.
;
BSTOP
;****************************************************************************
	MOVLW   B'00000000'     ; Put SCL, SDA line in output state
	bsf     STATUS,5        ;set the bank bit to 1
	movwf   TRISA           ; set the tri-state register
	bcf     STATUS,5        ;clear the bank bit to 0
;****************************************************************************
	BCF     PORTA,SDA          ; Return SDA to low
	BSF     PORTA,SCL          ; Set SCL high
	nop
	nop
	nop
	MOVLW   1               ; Ready error code 1
	BTFSS   PORTA,SCL          ; High?
	CALL    ERR             ; No, SCL locked low by device
	BSF     PORTA,SDA          ; SDA goes from low to high during SCL high
	MOVLW   4               ; Ready error code 4
	BTFSS   PORTA,SDA          ; High?
	CALL    ERR             ; No, SDA bus not release for STOP
	RETLW   0
;
;END SUB
;
;-----------------------------------------------------------------------------
;       Serial data send from PIC to serial EEPROM, bit-by-bit subroutine
;-----------------------------------------------------------------------------
;       Input   :       None
;       Output  :       To (DI) of serial EEPROM device
;-----------------------------------------------------------------------------
;
BITIN
	MOVLW   B'00001000'     ; Force SDA line as input
	bsf     STATUS,5        ;set the bank bit to 1
	movwf   TRISA           ; set the tri-state register
	bcf     STATUS,5        ;clear the bank bit to 0
	BSF     PORTA,SDA          ; Set SDA for input
	BCF     EEPROM,DI
	BSF     PORTA,SCL          ; Clock high
	MOVLW   1
	BTFSC   PORTA,SCL          ; Skip if SCL  is high
	GOTO    BIT1
	BTFSS   FLAG,ERR1       ; Remain as first error encountered
	MOVWF   ERCODE          ; Save error code
	BSF     FLAG,ERR1       ; Set error flag
BIT1
	BTFSC   PORTA,SDA          ; Read SDA pin
	BSF     EEPROM,DI       ; DI = 1
	NOP                     ; Delay
	BCF     PORTA,SCL          ; Return SCL to low     
	RETLW   0
;
;END SUB
; PAGE
;
;-----------------------------------------------------------------------------
;       Serial data receive from serial EEPROM to PIC, bit-by-bit subroutine
;-----------------------------------------------------------------------------
;       Input   :       EEPROM file
;       Output  :       From (DO) of serial EEPROM device to PIC
;-----------------------------------------------------------------------------
;
BITOUT
	MOVLW   B'00000000'     ; Set SDA, SCL as outputs
	bsf     STATUS,5        ;set the bank bit to 1
	movwf   TRISA           ; set the tri-state register
	bcf     STATUS,5        ;clear the bank bit to 0
	BTFSS   EEPROM,DO
	GOTO    BIT0
	BSF     PORTA,SDA          ; Output bit 0
	MOVLW   2
	BTFSC   PORTA,SDA          ; Check for error code 2
	GOTO    CLK1
	BTFSS   FLAG,ERR1       ; Remain as first error encountered
	MOVWF   ERCODE          ; Save error code
	BSF     FLAG,ERR1       ; Set error flag
	GOTO    CLK1            ; SDA locked low by device
;
BIT0
	BCF     PORTA,SDA          ; Output bit 0
	NOP                     ; Delay
	NOP
	NOP
CLK1
	BSF     PORTA,SCL
	MOVLW   1               ; Error code 1
	BTFSC   PORTA,SCL          ; SCL locked low?
	GOTO    BIT2            ; No.
	BTFSS   FLAG,ERR1       ; Yes.
	MOVWF   ERCODE          ; Save error code
	BSF     FLAG,ERR1       ; Set error flag
BIT2
	NOP
	NOP
	BCF     PORTA,SCL          ; Return SCL to low
	RETLW   0
;
;END SUB
; PAGE
;
;
;-----------------------------------------------------------------------------
;       RECEIVE         DATA    subroutine
;-----------------------------------------------------------------------------
;       Input   :       None
;       Output  :       RXBUF = Receive 8-bit data
;-----------------------------------------------------------------------------
;
RX
	MOVLW   .8              ; 8 bits of data
	MOVWF   COUNT
	CLRF    RXBUF
;
RXLP
	RLF     RXBUF, F        ; Shift data to buffer
	SKPC
	BCF     RXBUF,0         ; carry ---> f(0)
	SKPNC
	BSF     RXBUF,0
	CALL    BITIN
	BTFSC   EEPROM,DI
	BSF     RXBUF,0         ; Input bit =1
	DECFSZ  COUNT, F        ; 8 bits?
	GOTO    RXLP
	BSF     EEPROM,DO       ; Set acknowledge bit = 1
	CALL    BITOUT          ; to STOP further input
	RETLW   0
;
;END SUB
;
;-----------------------------------------------------------------------------
;       TRANSMIT        DATA    subroutine
;-----------------------------------------------------------------------------
;       Input   :       TXBUF
;       Output  :       Data X'mitted to EEPROM device
;-----------------------------------------------------------------------------
;
TX
	MOVLW   .8
	MOVWF   COUNT
;
TXLP
	BCF     EEPROM,DO       ; Shift data bit out.
	BTFSC   TXBUF,7         ; If shifted bit = 0, data bit = 0
	BSF     EEPROM,DO       ; Otherwise data bit = 1
	CALL    BITOUT          ; Serial data out
	RLF     TXBUF, F        ; Rotate TXBUF left
	SKPC                    ; f(6) ---> f(7)
	BCF     TXBUF,0         ; f(7) ---> carry
	SKPNC                   ; carry ---> f(0)
	BSF     TXBUF,0
	DECFSZ  COUNT, F        ; 8 bits done?
	GOTO    TXLP            ; No.
	CALL    BITIN           ; Read acknowledge bit
	MOVLW   3
	BTFSC   EEPROM,DI       ; Check for acknowledgement
	CALL    ERR             ; No acknowledge from device
	RETLW   0
;
;END SUB
; PAGE
;
;-----------------------------------------------------------------------------
;       BYTE-WRITE, write one byte to EEPROM device
;-----------------------------------------------------------------------------
;       Input   :       DATAO= data to be written
;                       ADDR    = destination address
;                       eepage  = 256 byte page address (1010xxx0)
;       Output  :       Data written to EEPROM device
;-----------------------------------------------------------------------------
;
WRBYTE
	rlf     eepage,0        ;mult page by 2 and load into w
	andlw   b'00001110'     ;mask out bits 0,4-7
	iorlw   b'10100000'     ;"add" in the rest of the bits
	MOVWF   TXBUF           ; to TX buffer
	CALL    BSTART          ; Generate START bit
	CALL    TX              ; Output page address
	MOVF    ADDR,W          ; Get WORD address
	MOVWF   TXBUF           ; into buffer
	CALL    TX              ; Output WORD address
	MOVF    DATAO,W         ; Move DATA 
	MOVWF   TXBUF           ; into buffer
	CALL    TX              ; Output DATA and detect acknowledgement
	CALL    BSTOP           ; Generate STOP bit
	return
;
;
;
;-----------------------------------------------------------------------------
;       BYTE-READ, read one byte from serial EEPROM device
;-----------------------------------------------------------------------------
;       Input   :       ADDR    = source address
;                       eepage  = 256 byte page address (1010xxx0)
;       Output  :       DATAI   = data read from serial EEPROM
;-----------------------------------------------------------------------------
;
RDBYTE
	rlf     eepage,0        ;mult page by 2 and load into w
	andlw   b'00001110'     ;mask out bits 0,4-7
	iorlw   b'10100000'     ;"add" in the rest of the bits
	MOVWF   TXBUF           ; into buffer (R/W = 0)
	CALL    BSTART          ; Generate START bit
	CALL    TX              ; Output page address. Check ACK.
	MOVF    ADDR,W          ; Get WORD address
	MOVWF   TXBUF
	CALL    TX              ; Output WORD address. Check ACK.
	CALL    BSTART          ; START READ (if only one device is
	rlf     eepage,0        ;mult page by 2 and load into w
	andlw   b'00001110'     ;mask out bits 0,4-7
	iorlw   b'10100000'     ;"add" in the rest of the bits
	MOVWF   TXBUF
	BSF     TXBUF,0         ; Specify READ mode (R/W = 1)
	CALL    TX              ; Output page address
	CALL    RX              ; READ in data and acknowledge
	CALL    BSTOP           ; Generate STOP bit
	MOVF    RXBUF,W         ; Save data from buffer
	MOVWF   DATAI           ; to DATAI file.
	return          ;end of rdbyte routine
	
	
	END     ;end of the skmf84.asm program


